package name.krestjaninoff.application.unused.session;

import java.util.Collections;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Map;

import org.apache.log4j.Logger;
import org.springframework.stereotype.Component;

import name.krestjaninoff.application.exception.SystemException;
import name.krestjaninoff.application.util.Algorithm;
import name.krestjaninoff.application.domain.User;


/**
 * User's sessions storage
 * 
 * @author Mikhail Krestjaninoff
 */
@Component("sessionManager")
public class SessionManagerImpl implements SessionManager {
	final Logger log = Logger.getLogger(SessionManagerImpl.class);

	/**
	 * Max idle time for user session
	 */
	final static Long MAX_IDLE_TIME = 30 * 60 * 1000L; 
	
	/**
	 * Sessions storage
	 */
	private Map<String, UserSession> sessions;
	
	
	/**
	 * Constructor
	 */
	public SessionManagerImpl() {
		this.sessions = Collections.synchronizedMap(new HashMap<String, UserSession>());
		
		this.startGarbageCollector();
	}

	
	/**
	 * Get user by session Id
	 */
	public synchronized User get(String sessionId) {
		UserSession userSession = sessions.get(sessionId);
		
		User user = userSession.getUser();
		userSession.updateAccessTime();
		
		return user;
	}
	
	/**
	 * Create user session
	 */
	public synchronized String save(User user) throws SystemException {
		String sessionId = Algorithm.uniqId();
		sessions.put(sessionId, new UserSession(sessionId, user));
		
		return sessionId;
	}
	
	/**
	 * Remove user session
	 */
	public synchronized void remove(String sessionId) {
		sessions.remove(sessionId);
	}
	
	
	/**
	 * Start garbage collector
	 */
	public void startGarbageCollector() {
		if (garbageCollector == null) {
			garbageCollector = new GarbageCollector();
			garbageCollector.setName("IdleSessionsGarbageCollector");
			garbageCollector.start();
		}
	}
	
	/**
	 * Stop garbage collector
	 * @throws InterruptedException
	 */
	public void stopGarbageCollector() throws InterruptedException {
		if (garbageCollector != null) {
			garbageCollector.finish();
			garbageCollector.join();
			garbageCollector = null;
		}
	}
	
	/**
	 * Garbage collector
	 * <br />
	 * Removes expired sessions
	 *
	 * @author Mikhail Krestjaninoff
	 */
	private class GarbageCollector extends Thread {
		boolean isActive = true;
		
		/**
		 * Remove expired sessions
		 */
		public void purgeSessions() {
			
			synchronized (sessions) {
                Iterator<UserSession> iter = sessions.values().iterator();
                while (iter.hasNext()) {
                	UserSession session = iter.next();
                    if (session.getIdleTime() > MAX_IDLE_TIME) {

                        log.info("Session with id " + session.getId()
                        		+ " expired and closed");
                        iter.remove();
                    }
                }
			}
		}
		
		/**
		 * Run garbage collector
		 */
		public void run() {
			log.info("User session garbage collector started");
			
			while (isActive) {
				try {
					synchronized (this) {
						wait(60 * 1000);
					}
				} catch (InterruptedException e) {
					isActive = false;
					break;
				}
				purgeSessions();
			}
		}

		/**
		 * Stop garbage collector
		 */
		public void finish() {
			isActive = false;
			
			synchronized (this) {
				notifyAll();
			}
			
			log.info("User session garbage collector stopped");
		}
	}
	
	private GarbageCollector garbageCollector;
}
